// ======================================================================
//
// TranslucentQueue.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "TranslucentQueue.h"

#include "GL_abstract.h"

// ======================================================================

GL_abstract                         *TranslucentQueue::gl;

TranslucentQueue::DrawFunction       TranslucentQueue::pointDraw;
TranslucentQueue::DrawFunction       TranslucentQueue::lineDraw;
TranslucentQueue::DrawFunction       TranslucentQueue::triangleDraw;

TranslucentQueue::Primitive         *TranslucentQueue::firstSorted;

const StaticMaterial                *TranslucentQueue::material;

int                                  TranslucentQueue::vertexBufferCount;
int                                  TranslucentQueue::primitiveBufferCount;

TVertex                              TranslucentQueue::vertexBuffer[VERTEX_BUFFER_SIZE];
TranslucentQueue::Primitive          TranslucentQueue::primitiveBuffer[PRIMITIVE_BUFFER_SIZE];

// ======================================================================

void TranslucentQueue::install(GL_abstract *newGl, DrawFunction newPointDraw, DrawFunction newLineDraw, DrawFunction newTriangleDraw)
{
	gl           = newGl;
	pointDraw    = newPointDraw;
	lineDraw     = newLineDraw;
	triangleDraw = newTriangleDraw;

	vertexBufferCount = 0;
	primitiveBufferCount = 0;
	firstSorted = NULL;
}

// ----------------------------------------------------------------------

void TranslucentQueue::remove(void)
{
}

// ----------------------------------------------------------------------

void TranslucentQueue::sort(Primitive *primitive)
{
	Primitive *back, *front;

	// search for insertion point
	for (back = NULL, front = firstSorted; front && primitive->sort < front->sort; back = front, front = front->next)
		;

	// hook into the list
	primitive->next = front;
	if (back)
		back->next = primitive;
	else
		firstSorted = primitive;
}

// ----------------------------------------------------------------------

void TranslucentQueue::addPoint(const TVertex *vertex)
{
	Primitive *primitive;

	// -qq- check for overflow
	primitive = primitiveBuffer + primitiveBufferCount++;

	primitive->material         = material;
	primitive->drawFunction     = pointDraw;
	primitive->vertices         = vertexBuffer + vertexBufferCount;
	primitive->numberOfVertices = 1;
	primitive->sort             = vertex->z;

	// -qq- check for overflow
	vertexBuffer[vertexBufferCount++] = *vertex;

	sort(primitive);
}

// ----------------------------------------------------------------------

void TranslucentQueue::addLine(const TVertex *vertex0, const TVertex *vertex1)
{
	Primitive *primitive;

	// -qq- check for overflow
	primitive = primitiveBuffer + primitiveBufferCount++;

	primitive->material         = material;
	primitive->drawFunction     = lineDraw;
	primitive->vertices         = vertexBuffer + vertexBufferCount;
	primitive->numberOfVertices = 2;
	primitive->sort             = min(vertex0->z, vertex1->z);

	// -qq- check for overflow
	vertexBuffer[vertexBufferCount++] = *vertex0;
	vertexBuffer[vertexBufferCount++] = *vertex1;

	sort(primitive);
}

// ----------------------------------------------------------------------

void TranslucentQueue::addTriangle(const TVertex *vertex0, const TVertex *vertex1, const TVertex *vertex2)
{
	Primitive *primitive;

	// -qq- check for overflow
	primitive = primitiveBuffer + primitiveBufferCount++;

	primitive->material         = material;
	primitive->drawFunction     = triangleDraw;
	primitive->vertices         = vertexBuffer + vertexBufferCount;
	primitive->numberOfVertices = 3;
	primitive->sort             = min(vertex0->z, min(vertex1->z, vertex2->z));

	// -qq- check for overflow
	vertexBuffer[vertexBufferCount++] = *vertex0;
	vertexBuffer[vertexBufferCount++] = *vertex1;
	vertexBuffer[vertexBufferCount++] = *vertex2;

	sort(primitive);
}

// ----------------------------------------------------------------------

void TranslucentQueue::draw(void)
{
	const StaticMaterial *oldMaterial;

	// draw all translucent polygons from back to front
	oldMaterial = NULL;

	while (firstSorted)
	{
		// handle material changes
		if (oldMaterial != firstSorted->material)
		{
			oldMaterial = firstSorted->material;
			gl->setMaterial(oldMaterial);
		}

		firstSorted->drawFunction(firstSorted->vertices, firstSorted->numberOfVertices);
		firstSorted = firstSorted->next;
	}

	// reset all the buffers for the next pass
	vertexBufferCount = 0;
	primitiveBufferCount = 0;
	firstSorted = NULL;
}

// ======================================================================

